import numpy as np
from numpy.testing import assert_array_almost_equal
from sanssouci.post_hoc_bounds import max_fp, min_tp, curve_max_fp, min_tdp
from sanssouci.post_hoc_bounds import curve_min_tdp, find_largest_region


def test_max_fp():
    p_values = np.linspace(1.e-6, 1 - 1.e-6, 100)
    p_values[:20] /= 10 ** 6

    # try with a scalar
    thresholds = np.array([1.e-4])
    assert max_fp(p_values, thresholds) == 80

    # try with a sequence
    thresholds = np.array([1.e-5, 1.e-4, 1.e-3])
    assert max_fp(p_values, thresholds) == 80

    # try with unordered sequence
    thresholds = np.array([1.e-3, 1.e-4, 1.e-5])
    assert max_fp(p_values, thresholds) == 80

    # try with pval = null value
    p_values = np.array([])
    thresholds = np.array([1.e-4])
    assert max_fp(p_values, thresholds) == 0  # return 0 here size of p_val ?

    # try with random sequence
    rng = np.random.RandomState(42)
    p_values = rng.rand(100)
    thresholds = np.array([1.e-4])
    assert max_fp(p_values, thresholds) == 100

    # corner case: empty vector
    p_values = ()
    thresholds = np.array([1.e-4])
    assert max_fp(p_values, thresholds) == 0
    

def test_min_tp():
    p_values = np.linspace(1.e-6, 1 - 1.e-6, 100)
    p_values[:20] /= 10 ** 6

    # try with a scalar
    thresholds = np.array([1.e-4])
    assert min_tp(p_values, thresholds) == 20

    # try with a sequence
    thresholds = np.array([1.e-5, 1.e-4, 1.e-3])
    assert min_tp(p_values, thresholds) == 20

    # try with unordered sequence
    thresholds = np.array([1.e-3, 1.e-4, 1.e-5])
    assert_array_almost_equal(min_tp(p_values, thresholds), (20, 20, 20))

    # try with pval = null value
    p_values = np.array([])
    thresholds = np.array([1.e-4])
    assert min_tp(p_values, thresholds) == 0  # size of pvalues return 0

    # try with random sequence
    rng = np.random.RandomState(42)
    p_values = rng.rand(100)
    thresholds = np.array([1.e-4])
    assert min_tp(p_values, thresholds) == 0


def test_min_tdp():
    p_values = np.linspace(1.e-6, 1 - 1.e-6, 100)
    p_values[:20] /= 10 ** 6

    # try with a scalar
    thresholds = np.array([1.e-4])
    assert min_tdp(p_values, thresholds) == 0.2

    # try with a sequence
    thresholds = np.array([1.e-5, 1.e-4, 1.e-3])
    assert min_tdp(p_values, thresholds) == 0.2

    # try with unordered sequence
    thresholds = np.array([1.e-3, 1.e-4, 1.e-5])
    assert_array_almost_equal(min_tdp(p_values, thresholds), (0.2, 0.2, 0.2))

    # try with pval = null value
    p_values = np.array([])
    thresholds = np.array([1.e-4])
    assert min_tdp(p_values, thresholds) == 0  # size of pvalues return 0

    # try with random sequence
    rng = np.random.RandomState(42)
    p_values = rng.rand(100)
    thresholds = np.array([1.e-4])
    assert min_tdp(p_values, thresholds) == 0


def test_curve_max_fp():
    p_values = np.linspace(1.e-6, 1 - 1.e-6, 100)
    p_values[:20] /= 10 ** 6

    # try with a scalar
    thresholds = np.array([1.e-4])
    curve_max_fp_ = curve_max_fp(p_values, thresholds)
    assert isinstance(curve_max_fp_, np.ndarray)
    assert all(curve_max_fp_ == np.append(np.zeros(20), np.arange(1, 81)))
    assert all(curve_max_fp_ <= max_fp(p_values, thresholds))
    assert len(curve_max_fp_) == len(p_values)
    assert all(curve_max_fp_ >= 0)
    assert curve_max_fp_[-1] == max_fp(p_values, thresholds)

    # try with a sequence
    thresholds = np.array([1.e-5, 1.e-4, 1.e-3])
    curve_max_fp_ = curve_max_fp(p_values, thresholds)
    assert isinstance(curve_max_fp_, np.ndarray)
    assert all(curve_max_fp_ == np.append(np.zeros(20), np.arange(1, 81)))
    assert all(curve_max_fp_ <= max_fp(p_values, thresholds))
    assert len(curve_max_fp_) == len(p_values)
    assert all(curve_max_fp_ >= 0)
    assert curve_max_fp_[-1] == max_fp(p_values, thresholds)

    # try with unordered sequence
    thresholds = np.array([1.e-3, 1.e-4, 1.e-5])
    curve_max_fp_ = curve_max_fp(p_values, thresholds)
    assert isinstance(curve_max_fp_, np.ndarray)
    assert all(curve_max_fp_ == np.append(np.zeros(20), np.arange(1, 81)))
    assert all(curve_max_fp_ <= max_fp(p_values, thresholds))
    assert len(curve_max_fp_) == len(p_values)
    assert all(curve_max_fp_ >= 0)
    assert curve_max_fp_[-1] == max_fp(p_values, thresholds)

    # try with random sequence
    rng = np.random.RandomState(42)
    p_values = rng.rand(100)
    thresholds = np.array([1.e-4])
    curve_max_fp_ = curve_max_fp(p_values, thresholds)
    assert isinstance(curve_max_fp_, np.ndarray)
    assert all(curve_max_fp_ == np.arange(1, 101))
    assert all(curve_max_fp_ <= max_fp(p_values, thresholds))
    assert len(curve_max_fp_) == len(p_values)
    assert all(curve_max_fp_ >= 0)
    assert curve_max_fp_[-1] == max_fp(p_values, thresholds)

    # try with size thresholds > p_values
    rng = np.random.RandomState(42)
    p_values = rng.rand(50)
    p_values[:10] /= 10 ** 6
    thresholds = rng.rand(100)
    curve_max_fp_ = curve_max_fp(p_values, thresholds)
    assert isinstance(curve_max_fp_, np.ndarray)
    assert all(curve_max_fp_ == np.append(np.zeros(10), np.arange(1, 41)))
    assert all(curve_max_fp_ <= max_fp(p_values, thresholds))
    assert len(curve_max_fp_) == len(p_values)
    assert all(curve_max_fp_ >= 0)
    assert curve_max_fp_[-1] == max_fp(p_values, thresholds)


def test_curve_min_tdp():
    p_values = np.linspace(1.e-6, 1 - 1.e-6, 100)
    p_values[:20] /= 10 ** 6
    # try with a scalar
    thresholds = np.array([1.e-4])
    curve_min_tdp_ = curve_min_tdp(p_values, thresholds)
    tns = 20 / np.arange(21, 101)
    assert isinstance(curve_min_tdp_, np.ndarray)
    assert all(curve_min_tdp_ == np.append(np.ones(20), tns))
    assert all(curve_min_tdp_ >= min_tdp(p_values, thresholds))
    assert len(curve_min_tdp_) == len(p_values)
    assert all(curve_min_tdp_ >= 0)
    assert curve_min_tdp_[-1] == min_tdp(p_values, thresholds)

    # try with a sequence
    thresholds = np.array([1.e-5, 1.e-4, 1.e-3])
    curve_min_tdp_ = curve_min_tdp(p_values, thresholds)
    tns = 20 / np.arange(21, 101)
    assert isinstance(curve_min_tdp_, np.ndarray)
    assert all(curve_min_tdp_ == np.append(np.ones(20), tns))
    assert all(curve_min_tdp_ >= min_tdp(p_values, thresholds))
    assert len(curve_min_tdp_) == len(p_values)
    assert all(curve_min_tdp_ >= 0)
    assert curve_min_tdp_[-1] == min_tdp(p_values, thresholds)


def test_find_largest_region():
    p_values = np.linspace(1.e-6, 1 - 1.e-6, 100)
    p_values[:20] /= 10 ** 6
    thresholds = np.array([1.e-5, 1.e-4, 1.e-3])
    TDP = 0.9
    region_size = find_largest_region(p_values, thresholds, TDP)
    assert region_size >= 0
    assert region_size == 21



import numpy as np
from sanssouci.row_welch import get_summary_stats
from sanssouci.row_welch import suff_welch_test
from sanssouci.row_welch import row_welch_tests
from numpy.testing import assert_array_almost_equal


def test_get_summary_stats():
    rng = np.random.RandomState(42)
    n = 20
    p = 50
    X = rng.randn(n, p)
    labels = rng.randint(2, size=n)
    summary = get_summary_stats(X, labels)

    assert isinstance(summary, dict)
    assert len(summary) == 2
    assert summary.keys() == set(labels)
    for key, values in summary.items():
        assert len(values) == 5
        assert isinstance(values, dict)
        assert values.keys() == {'sum', 'sum2', 'n', 'mean', 'sd'}

    assert summary[0]["n"] == n - summary[1]["n"]
    assert_array_almost_equal(summary[0]["sum"] + summary[1]["sum"],
                              np.sum(X, axis=0))
    assert_array_almost_equal(summary[0]["sum2"] + summary[1]["sum2"],
                              np.sum(X * X, axis=0))


def test_suff_welch_test():
    rng = np.random.RandomState(42)
    p = 50
    n = 20
    nx = 10
    ny = n - nx

    mx = rng.random_sample(p) - 1
    sx = rng.randn(p) * 0.5 + 1
    my = rng.random_sample(p)
    sy = rng.randn(p) * 0.1 + 2

    welch = suff_welch_test(mx, my, sx, sy, nx, ny)
    assert isinstance(welch, dict)
    assert len(welch) == 3
    assert welch.keys() == {'statistic', 'parameter', 'p_value'}
    for key, value in welch.items():
        assert len(value) == p
        assert isinstance(value, np.ndarray)


def test_row_welch_tests():
    rng = np.random.RandomState(42)
    n = 20
    p = 50
    X = rng.randn(n, p)
    labels = rng.randint(2, size=n)

    welch = row_welch_tests(X, labels)

    assert isinstance(welch, dict)
    assert len(welch) == 4
    assert welch.keys() == {'statistic', 'parameter', 'p_value', 'meanDiff'}
    for key, value in welch.items():
        assert len(value) == p
        assert isinstance(value, np.ndarray)



import numpy as np
from sanssouci.reference_families import inverse_linear_template
from sanssouci.reference_families import linear_template, beta_template
from sanssouci.reference_families import inverse_beta_template
from sanssouci.lambda_calibration import calibrate_jer, get_pivotal_stats
from scipy.stats import beta
from numpy.testing import assert_array_almost_equal


def test_linear_template():
    alpha = .05
    k = 5
    m = 10
    t = linear_template(alpha, k, m)
    assert_array_almost_equal(t, alpha * np.arange(1, k + 1) / m)
    assert isinstance(t, np.ndarray)
    assert len(t) == k


def test_inverse_linear_template():
    p = 20
    rng = np.random.RandomState(42)
    p0 = np.sort(rng.rand(p,))
    til = inverse_linear_template(p0, 2, p)
    assert (til >= p0).all()
    assert til.min() > 0.01
    assert til.max() < 10
    assert til.mean() > .9
    assert til.mean() < 5

    assert isinstance(til, np.ndarray)
    assert til.shape == p0.shape

    til_full_template = inverse_linear_template(p0, p, p)
    assert_array_almost_equal(til_full_template, p0)


def test_beta_template():
    alpha = .05
    k = 5
    m = 10
    t = beta_template(alpha, k, m)
    assert_array_almost_equal(t, beta.ppf(alpha, np.arange(1, k + 1),
                              np.array([m + 1] * k) - np.arange(1, k + 1)))
    assert isinstance(t, np.ndarray)
    assert len(t) == k


def test_inverse_beta_template():
    p = 20
    rng = np.random.RandomState(42)
    p0 = np.sort(rng.rand(p,))
    til = inverse_beta_template(p0, 2, p)
    assert (til >= p0).all()
    assert til.min() > 0.01
    assert til.max() <= 1
    assert til.mean() > .5
    assert til.mean() <= 1

    assert isinstance(til, np.ndarray)
    assert til.shape == p0.shape


def test_coherence_linear_template():
    """
    Testing whether linear_template and inverse_linear_template are actually
    inverse functions of one another. This also tests that calibration via
    dichotomy and pivotal stats are indeed equivalent.
    """
    alpha = 0.05
    B = 100
    p = 50
    seed = 42
    np.random.seed(seed)

    # Calibrate by dichotmy (calibrate_jer) using Simes template
    alphas = np.linspace(0, 1, 1000)
    simes_matrix = np.vstack([linear_template(alpha,
                                              p, p) for alpha in alphas])
    pval0 = np.random.uniform(size=(B, p))
    pval0 = np.sort(pval0, axis=1)
    calibrated_tpl = calibrate_jer(alpha, simes_matrix, pval0, k_max=p)

    # Calibrate using pivotal statistics

    piv_stat = get_pivotal_stats(pval0, K=p)
    lambda_quant = np.quantile(piv_stat, alpha)
    np.testing.assert_array_almost_equal(calibrated_tpl[-1],
                                         lambda_quant, decimal=3)



import numpy as np
from scipy.stats import beta


# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# REFERENCE FAMILIES
# R source code: https://github.com/pneuvial/sanssouci/
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


def linear_template(alpha, k, m):
    """
    Computes the linear template (Simes template)
    Parameters
    ----------
    alpha : float
        hyperparameter in [0, 1],
        to be chosen to satisfy Joint Error Rate Control.
    k : int
        number of rejection sets in linear template
    m : int
        number of hypotheses
    Returns
    -------
    template : array of shape (k,)
    """
    return alpha * np.arange(1, k + 1) / m


def inverse_linear_template(y, k, m):
    """
    Computes the inverse of the linear template (Simes template)
    Parameters
    ----------
    y : array of floats of shape (B, )
        values to apply template to
    k : int
        index of rejection set in linear template
    m : int
        number of hypotheses
    Returns
    -------
    estimated_template : array of same shape as y
    """

    return y * m / k


def beta_template(alpha, k, m):
    """
    Computes the Beta template
    Parameters
    ----------
    alpha : float
        hyperparameter in [0, 1],
        to be chosen to satisfy Joint Error Rate Control.
    k : int
        number of rejection sets in Beta template
    m : int
        number of hypotheses
    Returns
    -------
    template : array of shape (k,)
    """
    return beta.ppf(alpha, np.arange(1, k + 1),
                    np.array([m + 1] * k) - np.arange(1, k + 1))


def inverse_beta_template(y, k, m):
    """
    Computes the inverse of the Beta template
    Parameters
    ----------
    y : array of floats of shape (B, )
        values to apply template to
    k : int
        index of rejection set in beta template
    m : int
        number of hypotheses
    Returns
    -------
    estimated_template : array of same shape as y
    """
    return beta.cdf(y, k, m + 1 - k)


def shifted_linear_template(alpha, k, m, k_min):
    """
    Computes the shifted linear template (pARI template)
    Parameters
    ----------
    alpha : float
        hyperparameter in [0, 1],
        to be chosen to satisfy Joint Error Rate Control.
    k : int
        number of rejection sets in linear template
    m : int
        number of hypotheses
    k_min : int
        parameter that defines the shift of the template.
        The template is zero for k <= k_min.
    Returns
    -------
    template : array of shape (k,)
    """
    return alpha * np.array([
        max(0, (j - k_min) / (m - k_min))
        for j in range(1, k + 1)
    ])


def inverse_shifted_linear_template(y, k, m, k_min):
    """
    Computes the inverse of the shifted linear template (pARI template)
    Parameters
    ----------
    y : array of floats of shape (B, )
        values to apply template to
    k : int
        index of rejection set in linear template
    m : int
        number of hypotheses
    k_min :
        parameter that defines the shift of the template.
        The template is zero for k <= k_min.

    Returns
    -------
    estimated_template : array of same shape as y
    """
    if k - k_min <= 0:
        return np.array([np.inf] * y.shape[0])
    else:
        return y * (m - k_min) / (k - k_min)



import numpy as np
from numpy.testing import assert_array_almost_equal
from sanssouci.lambda_calibration import get_permuted_p_values
from sanssouci.lambda_calibration import get_pivotal_stats
from sanssouci.lambda_calibration import get_permuted_p_values_one_sample
from sanssouci.row_welch import row_welch_tests
from sanssouci.reference_families import inverse_linear_template
from scipy import stats


def test_get_permuted_p_values():
    rng = np.random.RandomState(42)
    n = 20
    p = 50
    X = rng.randn(n, p)
    B = 100
    categ = rng.randint(2, size=n)
    pvals = get_permuted_p_values(X, categ, B=B, row_test_fun=stats.ttest_ind)
    assert pvals.shape == (B, p)
    assert pvals.min() > 1.e-7
    assert pvals.max() <= 1
    assert np.sum(pvals < .1) < B * p * .12

    stats_ = [stats.ks_2samp, stats.bartlett, stats.ranksums, stats.kruskal]
    for stat in stats_:
        pvals = get_permuted_p_values(X, categ, B=B, row_test_fun=stat)
        assert pvals.min() > 1.e-7
        assert pvals.max() <= 1
        assert np.sum(pvals < .1) < B * p * .12

    pvals = get_permuted_p_values(X, categ, B=B, row_test_fun=row_welch_tests)
    assert pvals.shape == (B, p)
    assert pvals.min() > 1.e-7
    assert pvals.max() <= 1
    assert np.sum(pvals < .1) < B * p * .12


def test_get_permuted_p_values_one_sample():
    rng = np.random.RandomState(42)
    n = 20
    p = 50
    X = rng.randn(n, p)
    B = 100
    pvals = get_permuted_p_values_one_sample(X, B=B)
    assert pvals.shape == (B, p)
    assert (pvals == np.sort(pvals, axis=1)).all()
    assert pvals.min() > 1.e-7
    assert pvals.max() <= 1
    assert np.sum(pvals < .1) < B * p * .12


def test_get_pivotal_stat():
    B = 500
    p = 20
    rng = np.random.RandomState(42)
    p0 = np.sort(rng.rand(B, p), 1)

    piv_stat = get_pivotal_stats(p0,
                                 inverse_template=inverse_linear_template)
    assert piv_stat.shape == (B,)
    assert piv_stat.min() > 1.e-7
    assert piv_stat.max() < p0.max() * p
    assert isinstance(piv_stat, np.ndarray)
    tk_inv_all = np.array([inverse_linear_template(p0[:, i], i + 1, p)
                          for i in range(p)]).T
    assert_array_almost_equal(piv_stat, np.min(tk_inv_all[:, :p], axis=1))

    piv_stat = get_pivotal_stats(p0,
                                 inverse_template=inverse_linear_template, K=1)
    assert piv_stat.shape == (B,)
    assert piv_stat.min() > 1.e-7
    assert piv_stat.max() < p0.max() * p
    assert isinstance(piv_stat, np.ndarray)
    tk_inv_all = np.array([inverse_linear_template(p0[:, i], i + 1, p)
                          for i in range(p)]).T
    assert_array_almost_equal(piv_stat, np.min(tk_inv_all[:, :1], axis=1))

    piv_stat = get_pivotal_stats(p0, inverse_template=inverse_linear_template,
                                 K=p)
    assert piv_stat.shape == (B,)
    assert piv_stat.min() > 1.e-7
    assert piv_stat.max() < p0.max() * p
    assert isinstance(piv_stat, np.ndarray)
    tk_inv_all = np.array([inverse_linear_template(p0[:, i], i + 1, p)
                          for i in range(p)]).T
    assert_array_almost_equal(piv_stat, np.min(tk_inv_all[:, :p], axis=1))

    piv_statB = get_pivotal_stats(p0, inverse_template=inverse_linear_template,
                                  K=p)
    piv_statAll = get_pivotal_stats(p0,
                                    inverse_template=inverse_linear_template,
                                    K=-1)
    assert_array_almost_equal(piv_statB, piv_statAll)



from .row_welch import (
    get_summary_stats, suff_welch_test, row_welch_tests)
from .reference_families import (
    beta_template, inverse_beta_template,
    linear_template, inverse_linear_template,
    shifted_linear_template, inverse_shifted_linear_template)
from .post_hoc_bounds import (
    max_fp, min_tp, curve_max_fp, curve_min_tdp)
from .lambda_calibration import (
    calibrate_jer,
    get_permuted_p_values,
    get_permuted_p_values_one_sample,
    get_pivotal_stats,
    get_pivotal_stats_shifted
)


__all__ = [
    'get_summary_stats',
    'suff_welch_test',
    'row_welch_tests',
    'inverse_linear_template',
    'inverse_beta_template',
    'linear_template',
    'beta_template',
    'calibrate_jer',
    'get_permuted_p_values',
    'get_permuted_p_values_one_sample',
    'get_pivotal_stats',
    'get_pivotal_stats_shifted',
    'inverse_shifted_linear_template',
    'shifted_linear_template',
    'max_fp',
    'min_tp',
    'curve_max_fp',
    'curve_min_tdp',
]
